use std::collections::BTreeMap;

use serde::{Deserialize, Serialize};
use serde_with::{serde_as, DefaultOnNull, VecSkipError};
use time::OffsetDateTime;

use crate::util;

/// Bandcamp band object
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Band {
    pub id: u64,
    pub name: String,
    pub bandcamp_url: String,
    pub bio: String,
    /// Avatar image ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/<ID>_<SIZE>.jpg`
    pub bio_image_id: Option<u64>,
    /// Header image ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/<ID>_<SIZE>.jpg`
    pub header_image_id: Option<u64>,
    pub location: String,
    pub sites: Vec<Site>,
    pub discography: Vec<Discograph>,
}

/// Bandcamp album object
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Album {
    pub id: u64,
    pub title: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
    #[serde(rename = "type")]
    pub tralbum_type: TrAlbumType,
    pub tralbum_artist: String,
    pub band: BandInfo,
    pub bandcamp_url: String,
    pub credits: Option<String>,
    pub about: Option<String>,
    pub featured_track_id: u64,
    #[serde(with = "time::serde::timestamp")]
    pub release_date: OffsetDateTime,
    #[serde(default)]
    pub tags: Vec<Tag>,
    pub tracks: Vec<Track>,
}

/// Bandcamp track object
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Track {
    pub track_id: u64,
    pub title: String,
    pub album_id: u64,
    pub album_title: Option<String>,
    pub band_id: u64,
    pub band_name: Option<String>,
    /// Duration in seconds
    pub duration: f32,
    /// Map of audio formats and URLs.
    ///
    /// 128kbit/s MP3 is currently the only publicly available format.
    ///
    /// **Example:**
    ///
    /// `mp3-128` => `https://bandcamp.com/stream_redirect?enc=mp3-128&track_id=4173325157&ts=1686871135&t=a1f7904205e6caca970009ca6744ef0b512f353a`
    pub streaming_url: BTreeMap<String, String>,
    pub track_num: u16,
}

/// Discograph (album or track)
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Discograph {
    pub item_id: u64,
    pub item_type: DiscographType,
    pub title: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
    pub artist_name: Option<String>,
    pub band_id: u64,
    pub band_name: String,
    #[serde(with = "crate::serializer::date")]
    pub release_date: OffsetDateTime,
}

#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[serde(rename_all = "snake_case")]
pub enum DiscographType {
    Album,
    Track,
}

/// Band information
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct BandInfo {
    pub band_id: u64,
    pub name: String,
    pub bio: String,
    /// Avatar image ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/<ID>_<SIZE>.jpg`
    pub image_id: Option<u64>,
    pub location: String,
}

/// Album tag
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Tag {
    pub name: String,
    pub norm_name: String,
    pub isloc: bool,
}

/// Album tag
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Site {
    pub title: String,
    pub url: String,
}

/// Search result
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct SearchResult {
    #[serde_as(as = "VecSkipError<_>")]
    pub results: Vec<SearchItem>,
    pub time_ms: u32,
}

/// Search item
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
#[serde(tag = "type")]
pub enum SearchItem {
    #[serde(rename = "b")]
    Band(SearchBand),
    #[serde(rename = "a")]
    Album(SearchAlbum),
    #[serde(rename = "t")]
    Track(SearchTrack),
}

/// Band from search results
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct SearchBand {
    pub id: u64,
    pub name: String,
    /// Avatar image ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/<ID>_<SIZE>.jpg`
    pub img_id: Option<u64>,
    pub location: Option<String>,
    #[serde(default)]
    #[serde_as(deserialize_as = "DefaultOnNull")]
    pub tag_names: Vec<String>,
    pub url: String,
}

/// Album from search results
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct SearchAlbum {
    pub id: u64,
    pub name: String,
    pub band_id: u64,
    pub band_name: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
}

/// Track from search results
#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct SearchTrack {
    pub id: u64,
    pub name: String,
    pub band_id: u64,
    pub band_name: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
    pub album_id: Option<u64>,
    pub album_name: Option<String>,
}

/// Bandcamp feed containing featured items
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Feed {
    #[serde(with = "time::serde::timestamp")]
    pub feed_timestamp: OffsetDateTime,
    pub stories: FeedStories,
}

#[serde_as]
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct FeedStories {
    #[serde_as(as = "VecSkipError<_>")]
    pub featured: Vec<FeedItem>,
}

/// Bandcamp feed item
#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[non_exhaustive]
pub struct FeedItem {
    pub tralbum_id: u64,
    pub tralbum_type: TrAlbumType,
    pub title: String,
    pub artist: Option<String>,
    pub band_id: u64,
    pub band_name: String,
    pub band_url: String,
    /// Album art ID
    ///
    /// Image URL: `https://f4.bcbits.com/img/a<ID>_<SIZE>.jpg`
    pub art_id: Option<u64>,
    /// Feed item ID
    pub ntid: u64,
    #[serde(with = "time::serde::timestamp")]
    pub story_date: OffsetDateTime,
    pub story_type: StoryType,
    /// Used to fetch continuations
    pub story_token: String,
    #[serde(alias = "description")]
    pub desc: String,
    pub full_text: Option<String>,
    /// Some story entries link to an article on daily.bandcamp.com
    pub article_url: Option<String>,
    pub num_streamable_tracks: u16,
    #[serde(default)]
    pub tags: Vec<Tag>,
}

/// Feed story type
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[non_exhaustive]
pub enum StoryType {
    #[serde(rename = "nn")]
    Default,
    #[serde(rename = "ad")]
    Ad,
}

/// Track/Album type
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
pub enum TrAlbumType {
    #[serde(rename = "a")]
    Album,
    #[serde(rename = "t")]
    Track,
}

impl TrAlbumType {
    pub(crate) fn char(&self) -> char {
        match self {
            TrAlbumType::Album => 'a',
            TrAlbumType::Track => 't',
        }
    }
}

impl From<DiscographType> for TrAlbumType {
    fn from(value: DiscographType) -> Self {
        match value {
            DiscographType::Album => Self::Album,
            DiscographType::Track => Self::Track,
        }
    }
}

/// Search filter
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
#[non_exhaustive]
pub enum SearchFilter {
    #[serde(rename = "")]
    All,
    #[serde(rename = "b")]
    Bands,
    #[serde(rename = "a")]
    Albums,
    #[serde(rename = "t")]
    Tracks,
}

// Album UID

impl Album {
    /// Get a unified album ID.
    ///
    /// Unified album IDs allow for fetching Bandcamp albums using one string
    /// parameter. They are not an official part of Bandcamp.
    pub fn album_uid(&self) -> String {
        util::create_album_uid(self.band.band_id, self.tralbum_type, self.id)
    }
}

impl Discograph {
    /// Get a unified album ID.
    ///
    /// Unified album IDs allow for fetching Bandcamp albums using one string
    /// parameter. They are not an official part of Bandcamp.
    pub fn album_uid(&self) -> String {
        util::create_album_uid(self.band_id, self.item_type.into(), self.item_id)
    }
}

impl SearchAlbum {
    /// Get a unified album ID.
    ///
    /// Unified album IDs allow for fetching Bandcamp albums using one string
    /// parameter. They are not an official part of Bandcamp.
    pub fn album_uid(&self) -> String {
        util::create_album_uid(self.band_id, TrAlbumType::Album, self.id)
    }
}

impl FeedItem {
    /// Get a unified album ID.
    ///
    /// Unified album IDs allow for fetching Bandcamp albums using one string
    /// parameter. They are not an official part of Bandcamp.
    pub fn album_uid(&self) -> String {
        util::create_album_uid(self.band_id, self.tralbum_type, self.tralbum_id)
    }
}
